#!/usr/bin/env python
from __future__ import division

from sympy import collect, Symbol, diff, limit, factorial, cos, ccode
from sympy.abc import x,y,a
import numpy

class MyContext():
  def write(self, s):
    import sys
    sys.stdout.write(s)
    return

  def __getitem__(self, s):
    if s == 'store':
      return 'all'
    elif s == 'load':
      return ['locParent', 'locChild']
    return

def moveArray(context, localName, globalName, arraySize, globalOffsetName, numThreads = 1, blockNumber = 0, localBlockNumber = 0, isLoad = True, isCoalesced = True):
  # For a shared load (like G) on the CPU, which only loads once, we can eliminate 'idx'
  if not isCoalesced:
    if not numThreads == 1:
      raise RuntimeError('Non-coalesced moves must have numThreads == 1')
    if globalOffsetName:
      globalIdx = globalOffsetName+'+'
    else:
      globalIdx = ''
    localIdx  = ''
  else:
    if globalOffsetName:
      globalIdx = globalOffsetName+'+idx+'
    else:
      globalIdx = 'idx+'
    localIdx  = 'idx+'
  numMoves  = arraySize // numThreads
  remainder = arraySize % numThreads
  if isLoad:
    context.write('// Load %s from global memory into %s in shared memory\n' % (globalName, localName))
    for i in range(numMoves):
      globalOffset = i*numThreads + blockNumber*arraySize
      localOffset  = i*numThreads + localBlockNumber*arraySize
      if localName in context['load']:
        context.write('%s[%s%d] = %s[%s%d];\n' % (localName, localIdx, localOffset, globalName, globalIdx, globalOffset))
      else:
        context.write('%s[%s%d] = 0.5;\n' % (localName, localIdx, localOffset))
    if remainder:
      globalOffset = numMoves*numThreads + blockNumber*arraySize
      localOffset  = numMoves*numThreads + localBlockNumber*arraySize
      if localName in context['load']:
        context.write('if (idx < %d) %s[%s%d] = %s[%s%d];\n' % (remainder, localName, localIdx, localOffset, globalName, globalIdx, globalOffset))
      else:
        context.write('if (idx < %d) %s[%s%d] = 0.5;\n' % (remainder, localName, localIdx, localOffset))
  else:
    if not context['store'] == 'all': return ''
    context.write('// Store %s from shared memory into %s in global memory\n' % (localName, globalName))
    for i in range(numMoves):
      globalOffset = i*numThreads + blockNumber*arraySize
      localOffset  = i*numThreads + localBlockNumber*arraySize
      context.write('%s[%s%d] = %s[%s%d];\n' % (globalName, globalIdx, globalOffset, localName, localIdx, localOffset))
    if remainder:
      globalOffset = numMoves*numThreads + blockNumber*arraySize
      localOffset  = numMoves*numThreads + localBlockNumber*arraySize
      context.write('if (idx < %d) %s[%s%d] = %s[%s%d];\n' % (remainder, globalName, globalIdx, globalOffset, localName, localIdx, localOffset))
  return ''

# Series is broken
#f = 1/(x+y)
#print f.series(x, 0, 5)
#print f.series(x, 1, 5)

#t = newCenter - center
#array[0] = 1.0
#for i in range(1, numTerms):
#  array[i*numTerms] = (-1.0/i) * pow(-t, i) # First term is logarithmic, so treat separately
#  for j in range(1, i+1)
#    array[i*numTerms+j] = pow(-t, i-j) * getBinomial(i-1, j-1)

# f(x) = f(0) + (x - 0) f'(0) + 1/2! (x - 0)^2 f''(0) + ...
#      = [f(0) f'(0) 1/2! f''(0)]
#      = [m_0 m_1 m_2]
#
# f(x) = f(a) + (x - a) f'(a) + 1/2! (x - a)^2 f''(a) + ...
#      = [f(a) f'(a) 1/2! f''(a)]
#      = [n_0 n_1 n_2]
#      = f(a) - a f'(a) + 1/2! a^2 f''(a)
#      + (x - 0) (f'(a) - 2/2! a f''(a))
#      + (x - 0)^2 1/2! f''(a)
#      = [(n_0 - a n_1 + a^2/2 n_2) (n_1 - a n_2) n_2]
#
# So
# / 1 -a a^2/2 \ n_0   m'_0
# | 0  1   -a  | n_1 = m'_1
# \ 0  0    1  / n_2   m'_2

# Start with a monomial
# y = x + a
dim = 1
order = 5
c = [Symbol('c'+str(i)) for i in range(order)]
g = sum([c[i]*(x - a)**i/factorial(i) for i in range(order)])
print g
g = collect(g.expand(), x)
print g
M = []
for o in range(order):
  exp = g.diff(x, o).limit(x, 0)
  print exp
  M.append([exp.diff(c[p]) for p in range(order)])
print M

shift = 0.1
# Test that shifted series is close to re-expansion
sOrig = cos(x).series(x, 0, order)
print sOrig
#sShift = cos(x).series(x, 0.1, order)
sShift = cos(x + shift).series(x, 0, order)
print sShift
# Convert original series to a vector
vOrig = [float(sOrig.diff(x, o).limit(x, 0)) for o in range(order)]
print vOrig
# Convert shifted series to a vector
vShift = [float(sShift.diff(x, o).limit(x, 0)) for o in range(order)]
print vShift
# Produce M
Mn = numpy.zeros((order, order))
for o in range(order):
  exp = g.diff(x, o).limit(x, 0).limit(a, -shift)
  for p in range(order):
    Mn[o,p] = exp.diff(c[p])
print Mn
# Shift series
vShiftComp = numpy.dot(Mn, vOrig)
print vShiftComp
print vShift - vShiftComp
assert(numpy.linalg.norm(vShift - vShiftComp, 2) < 1.0e-2)

# Output an OpenCL code for this operation
#   1) Fully unrolled using shift 't'
#   2) Loops using shift 't'
#   3) Substitute 'level' for 't'
#
# Given M, let each thread do matvec
numThreads = 16
print '__kernel void M2M(__global float a, __global const float *child, __global float *parent) {'
print '  int   parentOffset = 0;'
print '  int   childOffset  = 0;'
print '  int   tidx         = get_local_id(0);'
print '  __local float locParent[%d];' % (order*numThreads)
print '  __local float locChild[%d];\n' % (order*numThreads*2**dim)
moveArray(MyContext(), 'locChild', 'child', order*numThreads*2**dim, 'childOffset', numThreads = numThreads, isLoad = True)
print '\n'.join([('  locParent[tidx*%d+%d] = ' % (order,o))+' + '.join(['(%s)*locChild[%d];' % (ccode(M[o][p]), p) for p in range(order)]) for o in range(order)])
moveArray(MyContext(), 'locParent', 'parent', order*numThreads, 'parentOffset', numThreads = numThreads, isLoad = False)
print '}'
# Output flop and byte count for kernel
# Quantify accuracy
